import importlib
from typing import Any, Dict, List, Union
from abc import ABC, abstractmethod
from functools import partial

if importlib.util.find_spec("jax") is not None:
    import jax
    from jax import random
    import jax.numpy as jnp
    import numpy as np
    from jax.lib import xla_bridge


def check_jax_device():
    try:
        print("Checking device used by JAX:")
        print(f"\tAvailable devices: {jax.devices()}")
        print(f"\tPlatform: {xla_bridge.get_backend().platform}")
    except Exception as e:
        print(f"Error while checking JAX device: {e}")


def nest_for_array(func):
    """Decorator to allow a function to be applied to nested arrays.

    Args:
        func (function): the function to decorate

    Returns:
        function: the decorated function
    """

    def wrapper(arr, *args, **kwargs):
        if isinstance(arr, jnp.ndarray):
            return func(arr, *args, **kwargs)
        elif isinstance(arr, dict):
            if "key_random" in kwargs:
                key_random = kwargs["key_random"]
                del kwargs["key_random"]
                for key, value in arr.items():
                    key_random, subkey = random.split(key_random)
                    arr[key] = wrapper(value, *args, key_random=subkey, **kwargs)
            else:
                for key, value in arr.items():
                    arr[key] = wrapper(value, *args, **kwargs)
            return arr
        elif isinstance(arr, list):
            if "key_random" in kwargs:
                key_random = kwargs["key_random"]
                del kwargs["key_random"]
                for idx, value in enumerate(arr):
                    key_random, subkey = random.split(key_random)
                    arr[idx] = wrapper(value, *args, key_random=subkey, **kwargs)
            else:
                for idx, value in enumerate(arr):
                    arr[idx] = wrapper(value, *args, **kwargs)
            return arr
        else:
            raise ValueError(f"Unknown type for array: {type(arr)}")

    return wrapper



import os
from typing import Dict, Type, Any, Tuple, Union
import random
import sys

from tbutils.once import one_time_printing


def try_get_seed(config: Dict, seed_max: int = 10000) -> int:
    """Will try to extract the seed from the config, or return a random one if not found

    Args:
        config (Dict): the run config

    Returns:
        int: the seed
    """
    try:
        seed = config["seed"]
        if not isinstance(seed, int):
            seed = random.randint(0, seed_max - 1)
            one_time_printing(
                f"[WARNING] The seed is not an integer, using a random seed instead: {seed}"
            )
    except KeyError:
        seed = random.randint(0, seed_max - 1)
    return seed


def set_seed(seed : int):
    """Set the seed for reproducibility. 
    This function will set the seed for Python, Numpy, PyTorch and TensorFlow, but only if the corresponding library is already imported.

    Args:
        seed (int): the seed to set
    """
    # Python
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    
    # Numpy
    if "numpy" in sys.modules:
        import numpy as np
        np.random.seed(0)
    
    # PyTorch
    if "torch" in sys.modules:
        import torch
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        # This flag only allows cudnn algorithms that are determinestic unlike .benchmark
        torch.backends.cudnn.deterministic = True

        # this flag enables cudnn for some operations such as conv layers and RNNs,
        # which can yield a significant speedup.
        torch.backends.cudnn.enabled = False

        # This flag enables the cudnn auto-tuner that finds the best algorithm to use
        # for a particular configuration. (this mode is good whenever input sizes do not vary)
        torch.backends.cudnn.benchmark = False    

    # TensorFlow
    if "tensorflow" in sys.modules:
        import tensorflow as tf
        tf.random.set_seed(seed)


from collections import defaultdict
import time
from typing import Any, Callable, Dict, Union


def timeit(func: Callable[..., Any]) -> Callable[..., Union[Any, float]]:
    """A wrapper function to return the result of the function and the time it took to execute it."""

    def time_measured_func(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        return result, end_time - start_time

    return time_measured_func


TOTAL_KEYWORD = "total"


class RuntimeMeter:
    """A context manager class to measure the time of various stages of the code take.

    with RuntimeMeter("train") as rm:
        # Training code
        #  ...
    with RuntimeMeter("eval") as rm:
        # Evaluation code
        #  ...
    with RuntimeMeter("train") as rm:   # second training phase, but the RuntimeMeter will add the time to the first training phase
        # Training code
        #  ...

    training_time = RuntimeMeter.get_runtime("train")
    eval_time = RuntimeMeter.get_runtime("eval")
    """

    stage_name_to_cum_runtime: Dict[str, float] = defaultdict(lambda: 0)
    stage_name_to_last_runtime: Dict[str, float] = defaultdict(lambda: None)
    stage_name_to_num_calls: Dict[str, int] = defaultdict(lambda: 0)

    @staticmethod
    def get_stage_runtime(stage_name: str) -> float:
        f"""Return the cumulative time taken by the stage.
        If the stage_name is {TOTAL_KEYWORD}, it will return the total time taken by all stages.
        If the stage_name is not found, it will return 0.

        Args:
            stage_name (str): the name of the stage, as it was used in the context manager.

        Returns:
            float: the cumulative time taken by the stage.
        """
        if stage_name == TOTAL_KEYWORD:
            return sum(RuntimeMeter.stage_name_to_cum_runtime.values())
        elif stage_name not in RuntimeMeter.stage_name_to_cum_runtime:
            return 0
        else:
            return RuntimeMeter.stage_name_to_cum_runtime[stage_name]

    @staticmethod
    def get_averaged_stage_runtime(stage_name: str) -> float:
        """Return the average time taken by the stage.

        Args:
            stage_name (str): the name of the stage, as it was used in the context manager.

        Returns:
            float: the average time taken by the stage.
        """
        if stage_name not in RuntimeMeter.stage_name_to_cum_runtime:
            return 0
        return (
            RuntimeMeter.stage_name_to_cum_runtime[stage_name]
            / RuntimeMeter.stage_name_to_num_calls[stage_name]
        )

    @staticmethod
    def get_last_stage_runtime(stage_name: str) -> float:
        """Return the time taken by the last call to the stage.

        Args:
            stage_name (str): the name of the stage, as it was used in the context manager.

        Returns:
            float: the time taken by the last call to the stage.
        """
        return RuntimeMeter.stage_name_to_last_runtime[stage_name]

    @staticmethod
    def get_runtimes() -> Dict[str, float]:
        """Return a dictionnary mapping the stage names to the cumulative time taken by the stage.

        Returns:
            Dict[str, float]: the dictionnary mapping the stage names to the cumulative time taken by the stage.
        """
        return dict(RuntimeMeter.stage_name_to_cum_runtime)

    @staticmethod
    def get_averaged_runtimes() -> Dict[str, float]:
        """Return a dictionnary mapping the stage names to the average time taken by the stage.

        Returns:
            Dict[str, float]: the dictionnary mapping the stage names to the average time taken by the stage.
        """
        return {
            stage_name: RuntimeMeter.get_averaged_stage_runtime(stage_name)
            for stage_name in RuntimeMeter.stage_name_to_cum_runtime
        }

    @staticmethod
    def get_last_runtimes() -> Dict[str, float]:
        """Return a dictionnary mapping the stage names to the time taken by the last call to the stage.

        Returns:
            Dict[str, float]: the dictionnary mapping the stage names to the time taken by the last call to the stage.
        """
        return dict(RuntimeMeter.stage_name_to_last_runtime)

    @staticmethod
    def get_total_runtime() -> float:
        """Return the total time taken by all stages.

        Returns:
            float: the total time taken by all stages.
        """
        return sum(RuntimeMeter.stage_name_to_cum_runtime.values())

    def __init__(self, stage_name: str, n_calls: int = 1):
        """Initialize the RuntimeMeter.

        Args:
            stage_name (str): a string identifying the stage.
            n_calls (int, optional): the number of calls to the stage. Defaults to 1.
        """
        assert stage_name != TOTAL_KEYWORD, (
            f"'{TOTAL_KEYWORD}' is a reserved keyword for the total time taken by all stages. "
            "If you want to measure the total time, use RuntimeMeter.get_total_runtime()."
        )
        self.stage_name = stage_name
        self.n_calls = n_calls

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stage_name_to_cum_runtime[self.stage_name] += time.time() - self.start_time
        self.stage_name_to_last_runtime[self.stage_name] = time.time() - self.start_time
        self.stage_name_to_num_calls[self.stage_name] += self.n_calls


def get_runtime_metrics():
    """Return the metrics of the runtimes.

    Returns:
        Dict[str, float]: a dictionnary mapping the stage names to the cumulative and averaged time taken by the stage.
    """
    dict_runtime_metrics = {}
    for stage_name in RuntimeMeter.stage_name_to_cum_runtime:
        dict_runtime_metrics[f"runtime/{stage_name}"] = RuntimeMeter.get_stage_runtime(
            stage_name
        )
        dict_runtime_metrics[f"runtime/{stage_name}_avg"] = (
            RuntimeMeter.get_averaged_stage_runtime(stage_name)
        )
        time_last_call = RuntimeMeter.get_last_stage_runtime(stage_name)
        if time_last_call is not None:
            dict_runtime_metrics[f"runtime/{stage_name}_last"] = time_last_call
    return dict_runtime_metrics


if __name__ == "__main__":
    import time
    import random

    def foo():
        time.sleep(0.1)

    def bar():
        time.sleep(0.2)

    for _ in range(3):
        with RuntimeMeter("foo"):
            foo()
        with RuntimeMeter("bar"):
            bar()

    print(RuntimeMeter.get_stage_runtime("foo"))
    print(RuntimeMeter.get_stage_runtime("bar"))
    print(RuntimeMeter.get_stage_runtime("total"))



import importlib
from typing import Any, Dict, List, Union
from abc import ABC, abstractmethod
from functools import partial

import os
import hydra
from omegaconf import OmegaConf


def merge_container(*containers):
    """Merge containers of the same type (list, dict, tuple) into one container."""
    containers = [OmegaConf.to_container(container) for container in containers]
    if all(isinstance(container, list) for container in containers):
        return [item for container in containers for item in container]
    elif all(isinstance(container, dict) for container in containers):
        return {
            key: value for container in containers for key, value in container.items()
        }
    elif all(isinstance(container, tuple) for container in containers):
        return tuple(item for container in containers for item in container)
    else:
        raise ValueError(
            f"All containers should be of the same type, but got {[type(container) for container in containers]}"
        )


def get_env_variable(name: str):
    """Get the value of an environment variable."""
    return os.environ.get(name)


def register_hydra_resolvers():
    """Register the custom Hydra resolvers.

    Example usage :
    ```yaml
    # configs/default.yaml
    key1: value1
    key2: ${eval:'${key1} /2'}
    ```
    """
    OmegaConf.register_new_resolver("merge", merge_container)
    OmegaConf.register_new_resolver("eval", eval)
    OmegaConf.register_new_resolver("env_variable", get_env_variable)


def instantiate_class(**kwargs) -> Any:
    """Instantiate a class from a dictionnary that contains a key "class_string" with the format "path.to.module:ClassName"
    and that contains other keys that will be passed as arguments to the class constructor

    Args:
        config (dict): the configuration dictionnary
        **kwargs: additional arguments to pass to the class constructor

    Returns:
        Any: the instantiated class
    """
    assert (
        "class_string" in kwargs
    ), "The class_string should be specified in the config"
    class_string: str = kwargs["class_string"]
    module_name, class_name = class_string.split(":")
    module = importlib.import_module(module_name)
    Class = getattr(module, class_name)
    object_config = kwargs.copy()
    object_config.pop("class_string")
    return Class(**object_config)



import sys
import platform
import importlib.util


def is_installed(package_name):
    return importlib.util.find_spec(package_name) is not None


def check_versions():
    python_version = sys.version.split(" ")[0]
    system_name = platform.system()
    machine = platform.machine()

    print("\n============== Checking Packages & System Info ================")
    print(f"OS: {system_name} ({machine})")
    print(f"Python: {python_version}")

    # NumPy
    if is_installed("numpy"):
        import numpy as np

        print(f"NumPy: {np.__version__}")

    # PyTorch
    if is_installed("torch"):
        import torch

        print(f"PyTorch: {torch.__version__}")
        print(f"\tCUDA Available: {torch.cuda.is_available()}")
        print(
            f"\tCUDA Version: {torch.version.cuda if torch.cuda.is_available() else 'N/A'}"
        )
        print(f"\tAvailable GPUs: {torch.cuda.device_count()}")
        for i in range(torch.cuda.device_count()):
            print(f"\t  - GPU {i}: {torch.cuda.get_device_name(i)}")

    # TensorFlow
    if is_installed("tensorflow"):
        import tensorflow as tf

        print(f"TensorFlow: {tf.__version__}")
        print(f"\tCUDA Available: {tf.test.is_built_with_cuda()}")
        print(f"\tAvailable GPUs: {len(tf.config.list_physical_devices('GPU'))}")
        for gpu in tf.config.list_physical_devices("GPU"):
            print(f"\t  - {gpu}")

    # JAX
    if is_installed("jax"):
        import jax
        from jax.lib import xla_bridge

        print(f"JAX: {jax.__version__}")
        print(f"\tAvailable Devices: {jax.devices()}")
        print(f"\tPlatform: {xla_bridge.get_backend().platform}")

    print("===============================================================")



from logging import INFO, WARNING, Logger
from typing import Callable

dict_printing_messages = set()


def one_time_printing(message : str):
    """Print a message only if it has not been printed before.

    Args:
        message (str): The message to print.
    """
    message = str(message)
    if message not in dict_printing_messages:
        print(message)
        dict_printing_messages.add(message)


dict_log_messages = set()


def one_time_log(logger: Logger, message: str, level=INFO, *args, **kwargs):
    """Log a <level> message only if it has not been logged before.

    Args:
        logger (Logger): The logger to log the message.
        message (str): The message to log.
        level: The level at which to log the message.
        args: The arguments to pass to the logger.
        kwargs: The keyword arguments to pass to the logger
    """
    message = str(message)
    if message not in dict_log_messages:
        logger.log(level, message, *args, **kwargs)
        dict_log_messages.add(message)


def one_time_warning(logger: Logger, message: str, *args, **kwargs):
    return one_time_log(logger, message, WARNING, *args, **kwargs)


def one_time_exec(func: Callable):
    """Decorator to make a function only execute once.

    Args:
        func (Callable): The function to decorate.
    """
    assert callable(func), f"Expected a callable, got {func}"
    def wrapper(*args, **kwargs):
        if not wrapper.has_executed:
            wrapper.has_executed = True
            return func(*args, **kwargs)

    wrapper.has_executed = False
    return wrapper



from setuptools import setup, find_packages

setup(
    name="tbutils",
    url="https://github.com/tboulet/mypackage", 
    author="Timothé Boulet",
    author_email="timothe.boulet0@gmail.com",
    
    packages=find_packages(),

    version="1.0",
    license="MIT",
    description="tbutils is a package of utility functions for python.",
    long_description=open('README.md').read(),      
    long_description_content_type="text/markdown",  
)


import sys
from typing import Any, Dict, Tuple, Union
import importlib
import numpy as np


def get_dict_flattened(d: Dict, parent_key="", sep="."):
    """Get a flattened version of a nested dictionary, where keys correspond to the path to the value.

    Args:
        d (Dict): The dictionary to be flattened.
        parent_key (str): The base key string (used in recursive calls).
        sep (str): Separator to use between keys.

    Returns:
        Dict: The flattened dictionary.
    """
    items = []
    for k, v in d.items():
        assert not sep in k, f"Separator {sep} is not allowed in keys"
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(get_dict_flattened(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def unflatten_dict(d: Dict, sep="."):
    """Unflatten a dictionary that was flattened using get_dict_flattened.

    Args:
        d (Dict): The dictionary to be unflattened.
        sep (str): Separator used between keys.

    Returns:
        Dict: The unflattened dictionary.
    """
    res = {}
    for key, value in d.items():
        keys = key.split(sep)
        current = res
        for k in keys[:-1]:
            if k not in current:
                current[k] = {}
            current = current[k]
        current[keys[-1]] = value
    return res

def get_shape(
    obj: Any,
    authorized_types: Tuple[type] = (np.ndarray, list, tuple, set, dict),
    assert_same_shape: bool = False,
) -> Tuple[int]:
    """Returns the shape of the object.
    If the object is a list, tuple, set or np.ndarray, it will return the shape of the object.
    If the object is not of the authorized types, it will return an empty tuple.

    This function makes the assumption that the object is a nested structure where sub-objects have the same shape.
    If assert_same_shape is True, it will assert that all sub-objects have the same shape.

    Args:
        object (Any): the object
        authorized_types (Tuple[type], optional): the authorized types for the object. Defaults to (np.ndarray, list, tuple, set).

    Returns:
        Tuple[int]: the shape of the object
    """
    if isinstance(obj, authorized_types):
        if len(obj) == 0:
            return (0,)
        if isinstance(obj, dict):
            obj = obj.values()
            shape = get_shape(list(obj), authorized_types, assert_same_shape)
            return f"Dict({len(obj)} keys to {shape})"
        first_shape = get_shape(next(iter(obj)), authorized_types, assert_same_shape)
        
        if assert_same_shape:
            sub_shapes = [get_shape(el, authorized_types, assert_same_shape) for el in obj]
            if any(shape != first_shape for shape in sub_shapes):
                raise AssertionError("All sub-objects must have the same shape, but got: ", sub_shapes)

        return (len(obj), *first_shape)
    
    return ()
